package lc.telemetry.historian.service;

import lc.esp.sdk.*;
import lc.esp.sdk.telemetry.TelemetryFrame;
import lc.esp.sdk.telemetry.TelemetrySymbol;
import lc.mecha.json.JSONArray;
import lc.mecha.json.JSONObject;
import lc.mecha.log.MechaLogger;
import lc.mecha.log.MechaLoggerFactory;
import lc.mecha.util.BasicallyDangerous;
import lc.mecha.util.StringAccumulatorV2;
import lc.mecha.util.UniversalJob;
import org.apache.http.HttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.BasicHttpEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.util.EntityUtils;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.Serializable;
import java.nio.charset.StandardCharsets;
import java.util.Map;
import java.util.Scanner;

/**
 * The historian service receives telemetry messages from the ESP
 * message fabric and relays those messages to an OpenTSDB server.
 *
 * @author Alex Leigh
 * @since mk18 (GIPSY DANGER)
 */
public class HistorianService extends BasicallyDangerous {
    public static final String TSDB_KEY_METRIC = "metric";
    public static final String TSDB_KEY_TIMESTAMP = "timestamp";
    public static final String TSDB_KEY_VALUE = "value";
    public static final String TSDB_KEY_TAGS = "tags";
    private static final MechaLogger logger = MechaLoggerFactory.getLogger(HistorianService.class);
    private final CloseableHttpClient httpClient = HttpClients.createDefault();

    public static void main(String[] args) {
        UniversalJob.banner(logger, "Historian mk1 (HORIZON BRAVE)");
        new HistorianService().run();
    }

    @Override
    public void runDangerously() throws Exception {
        ESPAddress telemDest = new ESPAddress("lc", "dev",
                "env", "mon", ESPAddressClass.TOPIC,
                ESPMessageClass.TELEMETRY);

        logger.info("Generated MQTT Address: {}", telemDest.toMQTTAddress());

        try (ESPClient esp = new ESPClient()) {
            esp.start();
            try (ESPSession session = esp.createSession()) {
                try (ESPConsumer consumer = session.createConsumer(telemDest)) {
                    //noinspection InfiniteLoopStatement
                    while (true) {
                        ESPMessage msg = consumer.receive();
                        publish(msg);
                    }
                }
            }
        }
    }

    /**
     * Publish the given telemetry message to OpenTSDB.
     */
    private void publish(ESPMessage msg) throws IOException {
        logger.info("Received message: {}", msg);
        HttpPost httpPost = new HttpPost("http://db4.leigh-co.com:4242/api/put");
        TelemetryFrame tf = new TelemetryFrame(msg);

        logger.info("Built frame: {}", tf);

        // opentsdb only supports seconds precision
        long seconds = tf.getTime().getEpochSecond();

        JSONArray arr = new JSONArray();

        for (TelemetrySymbol symbol : tf.getSymbols()) {
            JSONObject symJson = new JSONObject();

            StringAccumulatorV2 sa = new StringAccumulatorV2(".");
            sa.push(symbol.getName());
            // TODO: We need to enforce formatting here
            String uom = symbol.getUom();
            if (uom != null) {
                symJson.put(TSDB_KEY_METRIC, sa.asString() + "-" + uom);
            } else {
                symJson.put(TSDB_KEY_METRIC, sa.asString());
            }

            symJson.put(TSDB_KEY_TIMESTAMP, seconds);
            symJson.put(TSDB_KEY_VALUE, symbol.getValue());
            JSONObject tagsJson = new JSONObject();
            symJson.put(TSDB_KEY_TAGS, tagsJson);

            for (Map.Entry<String, Serializable> tag : tf.getTags().entrySet()) {
                tagsJson.put(tag.getKey(), tag.getValue());
            }

            arr.put(symJson);
        }

        // OpenTSDB does not support chunked requests!
        BasicHttpEntity requestEntity = new BasicHttpEntity();
        byte[] r = arr.toString().getBytes(StandardCharsets.UTF_8);
        logger.info("Built TSDB message: {}", arr.toString());

        requestEntity.setContent(new ByteArrayInputStream(r));
        requestEntity.setContentLength(r.length);
        requestEntity.setChunked(false);

        httpPost.setEntity(requestEntity);
        HttpResponse httpResponse = httpClient.execute(httpPost);

        // TSDB is very terse, it responds 204 on success
        if (httpResponse.getStatusLine().getStatusCode() != 204) {
            System.out.println(httpResponse.getStatusLine());
            Scanner sc = new Scanner(httpResponse.getEntity().getContent());
            while (sc.hasNext()) {
                System.out.println(sc.nextLine());
            }
        }

        EntityUtils.consume(httpResponse.getEntity());
    }
}
